"""Utilities to facilitate copying files and directories where one of the directories
could possibly be an S3 bucket.
This module requires that the AWS CLI is installed because "aws s3 sync" is used.
"""
import shutil
from pathlib import Path
import subprocess

def copy_dir_tree(from_dir, to_dir, content_type=None):
    """Recursively copies a directory to another location, deleting any files
    at that location that are not in the source directory.  'from_dir' and 'to_dir'
    directories can be S3 buckets + key prefix.
    If 'content_type' is not None, then the content-type of the files will be
    explicitly set to the passed in value.
    """
    # determine whether from and to directories are S3 directories
    from_is_s3 = (from_dir[:5] == 's3://')
    to_is_s3 = (to_dir[:5] == 's3://')
    
    # If neither directories are on S3, use a coventional copy tree command.
    # This will error if 'to_dir' exists, so first remove it if it is present.
    if from_is_s3 or to_is_s3:
        # use the AWS CLI S3 sync command.  Delete any destination files that
        # do not exist in the source.
        if content_type:
            subprocess.run(['aws', 's3', 'sync', from_dir, to_dir, '--delete', '--content-type', content_type])
        else:
            subprocess.run(['aws', 's3', 'sync', from_dir, to_dir, '--delete'])
    else:
        if Path(to_dir).exists():
            shutil.rmtree(to_dir)
        shutil.copytree(from_dir, to_dir)



"""Module used to configure exception and other application logging.
"""
import logging
import logging.handlers
import sys

def configure_logging(logging_module, log_file_path, log_level):
    """Configures the logging for an application.  Usage example:
           import logging
           import config_logging
           config_logging.configure_logging(logging, '/var/log/logfilename', 'INFO')

       Note that the 'log_level' parameter should be string matching a logging level
       attribute of the logging module: e.g. "INFO" or "DEBUG"
    """

    # Use the root logger for the application.

    # set the log level. Because we are setting this on the logger, it will apply
    # to all handlers (unless maybe you set a specific level on a handler?).
    logging_module.root.setLevel(getattr(logging, log_level))

    # create a rotating file handler
    fh = logging.handlers.RotatingFileHandler(log_file_path, maxBytes=200000, backupCount=5)

    # create formatter and add it to the handler
    formatter = logging_module.Formatter('%(asctime)s - %(levelname)s - %(module)s - %(message)s')
    fh.setFormatter(formatter)

    # filter out some records
    filtr = MyFilter()
    fh.addFilter(filtr)

    # create a handler that will print to console as well.
    console_h = logging_module.StreamHandler()
    console_h.setFormatter(formatter)
    console_h.addFilter(filtr)

    # add the handlers to the root logger
    logging_module.root.addHandler(fh)
    logging_module.root.addHandler(console_h)


class MyFilter(logging.Filter):
    """Class to restrict which logging records get logged.
    """

    def filter(self, record):

        # Don't record the records coming from the papermill execute module
        if record.module == 'execute':
            return False

        return True



# See: https://packaging.python.org/tutorials/packaging-projects/
# for how to install this package on PyPi.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# read version number from package
import bmonreporter
ver = bmonreporter.__version__

setuptools.setup(
    name="bmonreporter",
    version=ver,
    author="Alan Mitchell",
    author_email="tabb99@gmail.com",
    description="Creates Energy Reports from BMON Servers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alanmitchell/bmonreporter",
    packages=setuptools.find_packages(),
    scripts=['bin/bmonreporter'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'pyyaml>=5.1.1',
        'awscli>=1.16.264',
        'boto3>=1.10.0',
        'papermill>=1.2.0',
        'nteract-scrapbook>=0.3.1',
        'bmondata>=1.0.2',
        'plotly==3.10.0',
        'plotly_express==0.3.1',
        'jupyterthemes>=0.20.0',
    ]
)


"""Module to create the the HTML reports using Jupyter Notebooks as
the processing logic and final display.
"""

import sys
import logging
import tempfile
from pathlib import Path
from urllib.parse import urlparse
import subprocess
from typing import Iterable
from multiprocessing import Pool
from functools import partial

import boto3
import papermill as pm       # installed with: pip install papermill[s3], to include S3 IO features.
import scrapbook as sb       # install with: pip install nteract-scrapbook[s3], just in case S3 features used.
import bmondata

from bmonreporter.file_util import copy_dir_tree
import bmonreporter.config_logging

def create_reports(
        template_dir,      
        output_dir,
        bmon_urls,
        jup_theme_cmd,
        log_level,
        log_file_dir='bmon-reporter-logs/',
        cores=1,
    ):
    """Creates all of the reports for Organizations a Buildings across all specified BMON
    servers.

    Input Parameters:

    template_dir: directory or S3 bucket + prefix where notebook report templates 
        are stored.  Specify an S3 bucket and prefix by:  s3://bucket/prefix-to-templates
    output_dir: directory or S3 bucket + prefix where created reports are stored.
    bmon_urls: a list or iterable containing the base BMON Server URLs that should be 
        processed for creating reports.  e.g. ['https://bms.ahfc.us', 'https://bmon.analysisnorth.com']
    jup_theme_cmd: this is the Jupyter Theme command to run prior to generating reports.  This will
        set the formatting of the notebooks.  See: https://github.com/dunovank/jupyter-themes
    log_level: string indicating detail of logging to occur: DEBUG, INFO, WARNING, ERROR
    log_file_dir: (optional) directory or S3 bucket + prefix to store log files from report
        creation; defaults to 'bmon-report-logs' in current directory.
    cores: (optional) # of cores available to process this script.  Defaults to 1.
    """

    # set up logging
    # temporary directory for log files
    log_dir = tempfile.TemporaryDirectory()
    bmonreporter.config_logging.configure_logging(
        logging, 
        Path(log_dir.name) / 'bmonreporter.log', 
        log_level
    )

    try:
        # Run the Jupyter Themes command to get correct formatting of the notebook reports.
        subprocess.run(jup_theme_cmd, shell=True, check=True)

        # temporary directory for report templates
        templ_dir = tempfile.TemporaryDirectory()
        # copy the report templates into this directory
        copy_dir_tree(template_dir, templ_dir.name)

        # Loop through the BMON servers to process, but use the multiprocessing
        # module to do this in multiple processes.  To use multriprocessing you 
        # need to have a function with one parameter; create that with
        # functools.partial.
        server_func = partial(
            process_server, 
            template_dir=templ_dir.name,
            output_dir=output_dir 
            )
        cores_to_use = min(len(bmon_urls), cores)
        with Pool(cores_to_use) as p:
            p.map(server_func, bmon_urls)
        
    except:
        logging.exception('Error setting up reporter.')

    # copy the temporary logging directory to its final location
    copy_dir_tree(log_dir.name, log_file_dir, 'text/plain; charset=ISO-8859-15')

def process_server(server_url: str, template_dir: str, output_dir: str):
    """Create the reports for one BMON server with the base URL of 'server_url'.
    Pull report template notebooks from the 'template_dir' directory.
    Copy the reports to the directory specificed by 'output_dir', placed in a
    subdirectory named after the server_url.
    """
    # extract server domain for message labeling purposes
    server_domain = urlparse(server_url).netloc

    try:
        logging.info(f'Processing started for {server_domain}')
        
        # create a temporary directory to write reports
        rpt_dir = tempfile.TemporaryDirectory()
        rpt_path = Path(rpt_dir.name)
        
        # loop through all the buildings of the BMON site, running the building
        # templates on each.
        server = bmondata.Server(server_url)
        bldg_ids = [bldg['id'] for bldg in server.buildings()]
        run_report_set(
            server_url,
            'building_id',
            bldg_ids,
            Path(template_dir) / 'building',
            rpt_path / 'building',
            )

    except:
        logging.exception(f'Error processing server {server_domain}')

    # copy the report files to their final location
    dest_dir = str(Path(output_dir) / server_domain)

    # If the destination is s3, the Path concatenation above stripped out a /
    # that needs to be put back in.
    if dest_dir.startswith('s3:'):
        dest_dir = 's3://' + dest_dir[4:]
    
    copy_dir_tree(
        str(rpt_path), 
        dest_dir
    )

def run_report_set(
        server_url: str,
        param_name: str,
        param_values: Iterable,
        nb_template_path: Path,
        rpt_output_path: Path,
    ):
    """Creates a set of reports by cycling through a set of buildings or organizations
    and then cycling through a set of Jupyter notebook templates used to create the reports.
    Input Parameters:
        server_url: the full URL to BMON server that holds the data
        param_name: the name of building or organization parameter in the notebook template
        param_values: a list of values to cycle through for the notebook parameter.
        nb_template_path: a Path to the directory where the template report notebooks are stored.
        rpt_output_path: a Path to the directory where the final HTML reports are stored. A
            subdirectory for each parameter value will be created in this directory to hold all
            the reports for that particular building or organization.
    """
    # extract server domain for message labeling purposes
    server_domain = urlparse(server_url).netloc

    # create a temporary directory for scratch purposes, and make file
    # names inside that directory for the calculated report notebook and the HTML
    # files that is created from that notebook.
    scratch_dir = tempfile.TemporaryDirectory()
    out_nb_path = Path(scratch_dir.name) / 'report.ipynb'
    out_html_path = Path(scratch_dir.name) / 'report.html'

    # Keep track of how many reports completed and aborted
    completed_ct = 0
    aborted_ct = 0

    # Loop through all the parameter values (e.g. buildings or organizations)
    for param_val in param_values:
        if param_val != 2:
            continue
        for rpt_nb_path in nb_template_path.glob('*.ipynb'):

            try:
                pm.execute_notebook(
                    str(rpt_nb_path),
                    str(out_nb_path),
                    parameters = {'server_web_address': server_url, param_name: param_val},
                    kernel_name='python3',
                )

                # get the glued scraps from the notebook
                nb = sb.read_notebook(str(out_nb_path))
                scraps = nb.scraps.data_dict

                if 'hide' in scraps and scraps['hide'] == True:
                    # report is not available, probably due to lack of data
                    continue

                # convert the notebook to html. throw an error if one occurs.
                subprocess.run(f'jupyter nbconvert {out_nb_path} --no-input', shell=True, check=True)

                # move the resulting html report to the report directory
                # first create the destination file name and create the necessary
                # directories, if they don't exist.
                dest_name = Path(rpt_nb_path.name).with_suffix('.html')
                dest_path = rpt_output_path / str(param_val) / dest_name
                dest_path.parent.mkdir(parents=True, exist_ok=True)
                out_html_path.replace(dest_path)
                completed_ct += 1

            except pm.PapermillExecutionError as err:
                aborted_ct += 1
                if err.ename == 'RuntimeError':
                    # This error was raised intentionally to stop notebook execution.
                    # Just log an info message.
                    logging.info(f'Report aborted for server={server_domain}, {param_name}={param_val}, report={rpt_nb_path.name}: {err.evalue}')
                else:
                    logging.exception(f'Error processing server={server_domain}, {param_name}={param_val}, report={rpt_nb_path.name}')

            except:
                aborted_ct += 1
                logging.exception(f'Error processing server={server_domain}, {param_name}={param_val}, report={rpt_nb_path.name}')

    # log the number of completed and aborted reports
    logging.info(f'For server {server_domain}, report type {param_name}, {completed_ct} reports completed, {aborted_ct} reports aborted.')


def test():
    """Run the example configuration file as a test case.
    """
    import yaml
    config_file_path = '/home/tabb99/bmonreporter/bmonreporter/config_example.yaml'
    args = yaml.load(open(config_file_path), Loader=yaml.SafeLoader)
    create_reports(**args)

if __name__ == "__main__":
    test()